#include <stdio.h>
#include "parti.h"

void 
PARTI_scatter_append(sched, data, target, size)
SCHED *sched;			/* schedule */
char  *data,*target;		/* data arrays */
int   size;			/* size of each element */
{
  int count,i;
  char **blist,*buffer;
  char *ptr;
  void new_execute_scatter_append();

#ifdef DEBUG
  trace_enter("new_scatter_append");
#endif

  blist = wbuffer->blist;
  buffer = (char *) malloc(size*sched->ndata);
  count = 0;
  for(i=0;i<PARTI_numnodes;i++){
    blist[i] = &buffer[count];
    count += size * sched->sr->fetch_n[i];
  } 

  for(i=0,ptr=buffer;i<sched->ndata;i++,ptr+=size){ 
    bytecopy(ptr, &data[size*sched->perm[i]], size); 
  }  

  new_execute_scatter_append(sched->sr, blist, target, size);

  free(buffer);
#ifdef DEBUG
  trace_exit();
#endif
}

void
new_execute_scatter_append(pattern, buf_list, data, size)
struct com_pattern *pattern;
char **buf_list,*data;
int  size;
{
  static int call_count=0;
  int  i,j, p;
  int  *l;
  int  node;
  int  count,ccount;
  int  *msg_id;
  char *buffer, *ptr;
  char **wlist;
  char sync_buf[1];

#ifdef DEBUG
  trace_enter("new_execute_scatter");
#endif

  /* message buffer */
  msg_id = wbuffer->msg_id;
  if (size > sizeof(double))
    buffer = (char *) malloc(size*pattern->buffer_size);
  else
    buffer = (char *) wbuffer->buffer;

#ifdef	SYNC
  /* send data */
  for(count=0,i=0;i<PARTI_numnodes;i++) {
    if(pattern->fetch_n[i]) {
      PARTI_csend(SCATTER_EXECUTOR_MSG+call_count*MAXNODES+PARTI_mynode,
		  buf_list[i], size*pattern->fetch_n[i], i, PARTI_pid);
    }
  }

  /* recv data */
  for(ccount=0,i=0; i<PARTI_numnodes; i++) {
    if(pattern->send_n[i]) {
      PARTI_crecv(SCATTER_EXECUTOR_MSG+call_count*MAXNODES+i,
		  buffer, pattern->send_n[i]*size, i, PARTI_pid);
      for(j=0,ptr=buffer;j<pattern->send_n[i];j++,ccount++,ptr+=size){
          bytecopy(&data[ccount*size], ptr, size);
      }
    }
  }
#else
  /* set up the buffer list */
  wlist = wbuffer->wlist;
  for (p = 0, count = 0; p < PARTI_numnodes; p++) {
    wlist[p] = &data[count];
    count += (pattern->send_n[p]*size);
  }

  /* check if processor send messages to itself */
  if (pattern->fetch_n[PARTI_mynode]) {
    /* copy the buffer to target */
    bytecopy (wlist[PARTI_mynode], buf_list[PARTI_mynode], 
		pattern->fetch_n[PARTI_mynode]*size);
  }

  /* send and receive messages */
  for (i=1, count=0; i<PARTI_numnodes; i++){
    p = pattern->gcolor[i];
    /* call asynchronous receive funcations */
    if (pattern->send_n[p]) {
      msg_id[count++] =
      PARTI_irecv (FORCED_SCATTER_EXECUTOR_MSG+call_count*MAXNODES+p,
		   wlist[p], pattern->send_n[p]*size, p, PARTI_pid);
      PARTI_csend (SCATTER_SYNC+call_count*MAXNODES+PARTI_mynode,
		   sync_buf, 0, p, PARTI_pid);
    }

    /* send data */
    if (pattern->fetch_n[p]) {
      PARTI_crecv (SCATTER_SYNC+call_count*MAXNODES+p, sync_buf, 0, p, PARTI_pid);
      PARTI_bsend (FORCED_SCATTER_EXECUTOR_MSG+call_count*MAXNODES+PARTI_mynode,
		   buf_list[p], size*pattern->fetch_n[p], p, PARTI_pid);
    }
  }

  /* wait messages and store data */
  for (i = 1, count = 0; i < PARTI_numnodes; i++) {
    p = pattern->gcolor[i];
    if (pattern->send_n[p]) {
      PARTI_msgwait(msg_id[count]);
      count++;
    }
  }
#endif
  call_count = (++call_count > MAXCOUNT)? 0: call_count;

  if (size > sizeof(double))
    free(buffer);
#ifdef DEBUG
  trace_exit();
#endif
}

void
parti_scatter_append_(sched, data, target, size)
int  *sched;
char *data, *target;
int  *size;

{
  PARTI_scatter_append((SCHED *)get_sched_ptr(*sched), data, target, *size);
}

SCHED *
schedule_proc (proc,ndata,newndata,maxdim)
int *proc,ndata,maxdim;
int *newndata;			/* number of data after scatter_append */
{
  int *bl,*pl,**perms,**buckets,*bsize;
  int *GColor;
  int   i;
  int  count = 0;
  void	 bsort_proc();
  struct com_pattern *sr;
  struct com_pattern *schedule_gather_proc();
  SCHED *sched;
 
#ifdef DEBUG 
  trace_enter("schedule");
#endif
  sched		= (SCHED *) malloc(sizeof(SCHED));
  pl		= (int *) malloc(sizeof(int)*ndata);
  perms		= (int **) malloc(sizeof(int *)*PARTI_numnodes);
  bsize		= (int *) malloc(sizeof(int)*PARTI_numnodes);
  sched->global_refs = NULL;
  sched->local_refs = NULL;
  sched->type	= NULL;
  bsort_proc(PARTI_numnodes,ndata,proc,bsize,perms,pl);
  sr = schedule_gather_proc (bsize,maxdim);

  GColor  = (int *) malloc(sizeof(int)*PARTI_numnodes);
  PairColor(GColor, PARTI_mynode);
  sr->gcolor    = GColor;

  sched->sr 	= sr;
  sched->perm 	= pl;
  sched->ndata 	= ndata;

  /* count the number of data */
  for (i = 0; i < PARTI_numnodes; i++)
    count += sched->sr->send_n[i];
  *newndata = count;

  free(perms);
#ifdef DEBUG
  trace_exit();
#endif
  return(sched);
}

/*
 * wrapper function
 */
int
schedule_proc_ (proc,ndata,newndata,maxdim)
int *proc,*ndata,*maxdim;
int *newndata;			/* number of data after scatter_append */

{
  SCHED	*sched;

  sched = schedule_proc (proc, *ndata, newndata, *maxdim);
  return (register_schedule (sched));
}

/****************************************************************/
/* in:                                                          */
/*   nbuckets                                                   */
/*   ndata                                                      */
/*   key[ndata]                                                 */
/* out:                                                         */
/*   bsize[nbuckets]                                            */
/*   *perms[nbuckets]                                           */
/****************************************************************/
void 
bsort_proc(nbuckets,ndata,key,bsize,perms,plist)
int nbuckets,ndata,*key,*bsize,**perms,*plist;
{
  int i,count,*bcount,b;

  /* find size of each bucket */
  for(i=0;i<nbuckets;i++) bsize[i] = 0;
  for(i=0;i<ndata;i++){
    if(key[i]>=nbuckets){
      fprintf(stderr,
       "[%d]Error in bsort: not enough buckets [i-%d, key-%d, ndata-%d]\n", 
         PARTI_mynode, i, key[i], ndata);
      exit();
    }
    bsize[key[i]]++;
  }

  /* allocate memory for buckets */
  count = 0;
  for(i=0;i<nbuckets;i++){
    perms[i] = &plist[count];
    count += bsize[i];
  }

  /* do bucket sort */
  bcount = (int *) malloc(sizeof(int)*nbuckets);
  for(i=0;i<nbuckets;i++) bcount[i] = 0;
  for(i=0;i<ndata;i++){
    b = key[i];
    perms[b][bcount[b]] = i;
    bcount[b]++;
  }
  free(bcount);
}

#define	 SEND_MATR	510000	/* message id for matrix transpose */
#define	 SEND_SCHED	520000	/* message id for gathering schedules */

struct com_pattern *
schedule_gather_proc (n,maxdim)
int *n;
int maxdim;
 
{
  int i,j,k,no;
  int *a,*wrk,n_msgs_to_send; /* used in 1st part */
  int *l,msg_size,node,*send_n,**send_list;  /* used in second part */
  struct com_pattern *out_struct;
  int n_msgs_to_recv;
  int size, f_size, s_size;
  int *msgid;
  static int sched_call = 0;

#ifdef DEBUG 
  trace_enter("schedule_gather"); 
#endif 

  /* find how many messages to be sent in exec phase */
  /* i.e. find the transpose of the original distributed matrix */
  send_n = (int *) malloc(sizeof(int)*PARTI_numnodes);
  msgid = wbuffer->msg_id;
  /* receive elements from other processors */
  for (i = 0, no = 0; i < PARTI_numnodes; i++) {
    if (i == PARTI_mynode)
      continue;
    msgid[no++] = PARTI_irecv(SEND_MATR+i, &send_n[i], 
					sizeof(int), i, PARTI_pid);
  }
  /* send the elements to other processors */
  for(i=1;i<PARTI_numnodes;i++){
    k = (PARTI_mynode+i) % PARTI_numnodes;
    PARTI_csend(SEND_MATR+PARTI_mynode, &n[k], sizeof(int), k, PARTI_pid);
  }  
  send_n[PARTI_mynode] = n[PARTI_mynode];
  /* wait until all asynchronous receives are done */
  for (i = 0; i < no; i++) {
    PARTI_msgwait (msgid[i]);
  }

  /* put pointers into structre to be passed out */
  out_struct = (struct com_pattern *) malloc(sizeof(struct com_pattern));
  out_struct->buffer_size = 0;
  f_size = s_size =  0;
  for(i=0;i<PARTI_numnodes;i++){
    if(out_struct->buffer_size<send_n[i]){
      out_struct->buffer_size = send_n[i];
    }
    s_size += send_n[i];
    f_size += n[i];
  }
  size = (s_size < f_size)? f_size: s_size; 

  if (wbuffer->max_size < (size*maxdim*sizeof(double))) {
    free(wbuffer->buffer); 
    wbuffer->buffer = (char *)malloc((size*maxdim)*sizeof(double)); 
    if (NULL == wbuffer->buffer)
      fprintf(stderr,"ERROR:not enough memory -parti working buffer\n"),exit(0);
    wbuffer->max_size = size * maxdim * sizeof(double);
  }
  n_msgs_to_recv = 0;
  n_msgs_to_send = 0;
  for (i = 0; i < PARTI_numnodes; i++) {
    if (n[i]) n_msgs_to_recv++;
    if (send_n[i]) n_msgs_to_send++;
  }

  out_struct->n_msgs_to_recv = n_msgs_to_recv;
  out_struct->n_msgs_to_send = n_msgs_to_send;
  out_struct->send_n = send_n;
  out_struct->send_list = NULL;
  out_struct->fetch_n = n;
  out_struct->fetch_list = NULL;
  out_struct->dimen = maxdim;
  out_struct->send_size = NULL;
  out_struct->fetch_size = NULL;
  out_struct->total_send_size = s_size;
  out_struct->total_fetch_size = f_size;

 sched_call++;
#ifdef DEBUG
  trace_exit();
#endif
 
  return(out_struct);
}  
